from jija import views

from aiohttp import web


class Router:
    def __init__(self, endpoints):
        self.__routes = self.__generate_routes(endpoints)

    @property
    def routes(self):
        return self.__routes

    @staticmethod
    def __generate_routes(endpoints):
        result = []
        for endpoint in endpoints:
            result.extend(endpoint.generate_routes())

        return result


class AbsEndpoint:
    def generate_routes(self, prefix=''):
        raise NotImplementedError()


class Endpoint(AbsEndpoint):
    def __init__(self, path, view):
        if not issubclass(view, views.View):
            raise AttributeError(f'view must be a subclass of "jija.views.View", got {type(view)}')

        self.__path = path
        self.__view = view

    def generate_routes(self, prefix=''):
        result = []
        for method in self.__view.get_methods():
            result.append(web.route(method, f'{prefix}{self.__path}', self.__view.construct))

        return result


class Include(AbsEndpoint):
    def __init__(self, path, endpoints):
        self.__path = path
        self.__endpoints = endpoints

    def generate_routes(self, prefix=''):
        result = []
        for endpoint in self.__endpoints:
            result.extend(endpoint.generate_routes(prefix))

        return result



from jija.commands.migrate import Migrate
from jija.commands.run import Run
from jija.commands.update import Update
from jija.commands.test import Test
from jija.commands.runprocess import RunProcess

COMMANDS = {
    'run': Run,
    'migrate': Migrate,
    'update': Update,
    'test': Test,
    'runprocess': RunProcess
}



from aiohttp import web


class View:
    methods = ('get', 'post', 'patch', 'put', 'delete')

    def __init__(self, request: web.Request, path_params: web.UrlMappingMatchInfo):
        self.__request = request
        self.__path_params = path_params

    @classmethod
    def get_methods(cls):
        view_methods = []
        for method in cls.methods:
            if hasattr(cls, method):
                view_methods.append(method)

        return view_methods

    @classmethod
    async def construct(cls, request: web.Request):
        view = cls(request, request.match_info)
        return await view.dispatch()

    @property
    def request(self):
        return self.__request

    async def dispatch(self):
        method = self.request.method.lower()
        handler = getattr(self, method)

        return await handler()

    async def serialize(self):
        pass



from jija import apps


class Config:
    __REQUIRED_ADDONS = set()
    __PREF = {}

    def __init__(self, **kwargs):
        self.__class__.__PREF = kwargs
        apps.Apps.config_init_callback(self.__class__)

    @classmethod
    async def freeze(cls):
        validated_data = await cls.validate(cls.__PREF)
        cls.set_values(validated_data)

    @classmethod
    async def validate(cls, values):
        validated_data = {}
        for name, value in values.items():
            name = name.upper()
            field = getattr(cls, name)
            validated_data[name.upper()] = await field.validate(value)

        return validated_data

    @classmethod
    def set_values(cls, validated_data):
        for name, value in validated_data.items():
            setattr(cls, name, value)

    @classmethod
    def base_app_update(cls, aiohttp_app):
        return aiohttp_app

    @classmethod
    def each_app_update(cls, aiohttp_app):
        return aiohttp_app

    @classmethod
    async def preflight(cls):
        pass



from jija.config import base, fields
from jija import drivers


class DriversConfig(base.Config):
    DOCS: drivers.DocsDriver = fields.ClassField(
        class_pattern=drivers.DocsDriver, default=drivers.DocsDriver)
    DATABASE: drivers.DatabaseDriver = fields.ClassField(
        class_pattern=drivers.DatabaseDriver, default=drivers.DatabaseDriver)

    def __init__(self, *, docs=None, database=None):
        super().__init__(docs=docs, database=database)



from aiohttp import web


from jija.command import Command
from jija.config import NetworkConfig


class RunProcess(Command):
    def handle(self):
        from jija.apps import Apps
        web.run_app(Apps.core.aiohttp_app, loop=self.loop, host=NetworkConfig.HOST, port=NetworkConfig.PORT)



import asyncio
import subprocess

from jija import config
from jija.command import Command
from jija import reloader


class Run(Command):
    def __init__(self):
        super().__init__()
        self.close_event = asyncio.Event()

    async def run_watcher(self):
        reloader_instance = reloader.Reloader(config.StructureConfig.PROJECT_PATH, self.close_event)
        await reloader_instance.wait()

    async def handle(self):
        asyncio.create_task(self.run_watcher())
        while True:
            runner = subprocess.Popen([config.StructureConfig.PYTHON_PATH, 'main.py', 'runprocess'])
            self.close_event.clear()
            await self.close_event.wait()
            runner.kill()
            print()



from setuptools import setup

setup(
    name='jija',
    version='0.0.8',
    description='',
    packages=[
        'jija',
        'jija.database',
        'jija.forms',
        'jija.commands',
        'jija.utils',
        'jija.middlewares',
        'jija.config',
        'jija.drivers',
        'jija.contrib.auth'
    ],
    author='Kain',
    author_email='kainedezz.2000@gmail.com',
    zip_safe=False,

    install_requires=[
        'aiohttp==3.8.1',
        'aiofile==3.8.1',
        'jija_orm==0.0.1',
        'cryptography',
        'aiohttp_session[secure]',
    ]
)



import asyncio
import aiofile

import hashlib
import os


class Reloader:
    def __init__(self, root_path, event: asyncio.Event):
        self.__root_path = '/app'
        # self.__root_path = root_path  # TODO
        self.__event = event

    async def wait(self):
        last_hash = None
        while True:
            try:
                new_hash = await self.__get_hash(self.__root_path)
                if last_hash is None:
                    last_hash = new_hash

                if last_hash != new_hash and not self.__event.is_set():
                    last_hash = new_hash
                    self.__event.set()
                    await asyncio.sleep(3)

                await asyncio.sleep(1)

            except Exception as exception:
                print('Reloader caught an exception but still work')
                print(exception)
                print()
                await asyncio.sleep(5)

    async def __get_hash(self, path, pre_hash=None):
        if not pre_hash:
            pre_hash = hashlib.sha1()

        files = os.listdir(path)

        for file in files:
            next_path = f'{path}/{file}'
            if os.path.isdir(next_path):
                pre_hash = await self.__get_hash(next_path, pre_hash)

            else:
                async with aiofile.async_open(next_path, 'rb') as buffer:
                    chunk = 0
                    while chunk != b'':
                        chunk = await buffer.read(1024)
                        pre_hash.update(chunk)

        if path == self.__root_path:
            return pre_hash.hexdigest()

        return pre_hash



from jija.config import base, fields


class DatabaseConfig(base.Config):
    DATABASE = fields.CharField()
    PASSWORD = fields.CharField()
    USER = fields.CharField(default='postgres')
    PORT = fields.IntegerField(default=5432)
    HOST = fields.CharField(default='localhost')

    def __init__(self, *, database, password, host=None, user=None, port=None):
        super().__init__(database=database, password=password, host=host, user=user, port=port)

    @classmethod
    async def preflight(cls):
        from jija import config
        await config.DriversConfig.DATABASE.preflight()
